//=============================================================================
// AssignParser.g - Based on AssignExprsTree, build a Tree for 
//                  Assignment Statements and Arithmetic Expressions
//                  including Function Invocation and 1-D Arrays
//                  Then use a tree-walker to walk the AST and emit
//                      code for an imaginary stack machine...
//                  Use of a variable or array on the LHS (as an L-value)
//                      constitutes declaring/defining it
//                  Semantics assume a stack machine with tagged entries
//                      Tags:  value, address, fcn-address
//
// Author:  Bary W Pollack
// Date:    Apr. 17, 2004 - 16:00
//
//=============================================================================

// Import the necessary classes
{
import java.io.*;
import java.util.*;
import antlr.debug.misc.ASTFrame;
}


//-----------------------------------------------------------------------------
// Define a Parser, call it AssignParse
//-----------------------------------------------------------------------------

class AssignParser extends Parser;

options {
    defaultErrorHandler = false;    // Don't generate parser error handlers
    buildAST = true;                // Turn on AST tree construction
}

tokens {
    ARRAY; CALL; PROGRAM; 
}


// Define some methods and variables to use in the generated parser
{
    // Define a main
    public static void main(String[] args) {
        // Use a try/catch block for parser exceptions
        try {
            // if we have at least one command-line argument
            if (args.length > 0 ) {
                System.out.println();
                System.out.println("AssignParser Parsing...");
                System.out.println();

                // For each directory/file specified on the command line
                for (int i = 0; i < args.length; i++)
                    doFile(new File(args[i]));  // parse it
            } else
                System.out.println("Usage: java AssignParser [ file | <directory name> ]...");

        } catch (Exception e) {
            System.out.println("Exception: " + e);
            e.printStackTrace(System.out);   // so we can get a stack trace
        }
    }


    // This method decides what action to take based on the
    // type of file we are looking at -- directory or source file
    public static void doFile(File f) throws Exception {
        // If this is a directory, walk each file/dir in that directory
        if (f.isDirectory()) {
            String files[] = f.list();
            for (int i = 0; i < files.length; i++)
                doFile(new File(f, files[i]));        
        // otherwise, if this is a source file, parse it
        } else if ((f.getName().length() >= 5) &&
                   f.getName().substring(f.getName().length() - 4).equals(".txt")) {
            ///System.out.println("  File: " + f.getAbsolutePath());  //FOO
            System.out.println("  File: " + f.getName());
            parseFile(new FileInputStream(f));
        }
    }

    // Here's where we do the real work...
    public static void parseFile(InputStream s) throws Exception {
        System.out.println();
        try {
            // Create a scanner that reads from the input stream passed to us
            AssignLexer lexer = new AssignLexer(s);

            // Create a parser that reads from the scanner
            AssignParser parser = new AssignParser(lexer);

            // Start parsing at the program rule
            parser.program();
            System.out.println();
            System.out.println("* Parsing Successful *");

            // Create an AST for the input expression
            pSymbols = new Symbol();  // initialize the symbol table
            AST tree = parser.getAST();
            System.out.println();
            System.out.println("Tree: " + tree.toStringTree());
            System.out.println();
            System.out.println("* Tree Construction Successful *");
            System.out.println();

            // Create an ASTFrame to show the tree structure in 2D
            if (true) {
                ASTFrame frame = new ASTFrame("Tree for: " + tree.toStringList(), tree);
                frame.setVisible(true);
            }

            // Walk the AST to create the semantics (evaluate the expression)
            pSymbols = new Symbol();  // initialize the symbol table
            AssignTreeWalker treeWalker = new AssignTreeWalker();
            treeWalker.program(tree);

            System.out.println();
            System.out.println("* Tree-Walk Successful *");
            System.out.println();

            // Display the Symbol Table
            if (true) {           // control display of symbol table  //FOO
                System.out.println();
                pSymbols.displaySymbols();
                System.out.println();
            }

            // Check to see if there were any errors
            if (nErrors == 0)
                System.out.println("NO SEMANTIC ERRORS DETECTED");
            else {
                System.out.println(nErrors + " SEMANTIC ERRORS DETECTED; "
                                   + "code generation will proceed anyway...");
            }
 
            // Walk the AST to create the semantics (evaluate the expression)
            System.out.println();
            System.out.println("* Begin Code Generation *");
            System.out.println();
            pSymbols = new Symbol();  // initialize the symbol table
            AssignGen treeWalker2 = new AssignGen();
            treeWalker2.program(tree);

            System.out.println();
            System.out.println("* SECOND Tree-Walk Successful *");
            System.out.println();

        } catch (Exception e) {
            System.out.println("Parser exception: " + e);
            System.out.println();
            e.printStackTrace();   // so we can get a stack trace

        } finally {
            if (true) {           // control display of symbol table  //FOO
                System.out.println();
                pSymbols.displaySymbols();
                System.out.println();
            }
            if (nErrors > 0) {
                System.out.println(nErrors + " ERRORS were detected during compilation");
            }
        }
    }

    public static Symbol pSymbols = null;       // This is the symbol table

    public static void sop(String s) {          //FOO
        if (true)                               //FOO
        System.out.println(s); 
    }

    public static void error() { ++nErrors; }

    private static int nErrors = 0;
}

// ==================================================================

program
  : programBody 
      { sop("program");
        #program = #([PROGRAM, "PROGRAM"], program); 
      }
    // end-of-file
  ;

programBody
  : "begin"!
       statementList
    "end"!
  ;

statement
  : (IDENT (LPAREN|SEMI))=> procedureCallStatement
  | assignmentStatement
  ;

statementList
  : statement statementList
  | // nothing
  ;

assignmentStatement
  { String varx = "who"; }
  : varx=variableReference EQUALS^ expression SEMI!
      { sop("assignmentStatement: " + varx); sop(""); }
  ;

procedureCallStatement
  : ident:IDENT actualParameters SEMI!
      { System.out.println("ident=  " + ident.getText());    //FOO
        sop("procedureCallStatement: " + ident.getText()); sop("");
        #procedureCallStatement = #([CALL, "CALL"], procedureCallStatement);
      }
  ;

actualParameters
  : LPAREN! (expression (COMMA! expression)*)? RPAREN!
  ;

procedureCall
  : ident:IDENT actualParameters
      { System.out.println("ident=  " + ident.getText());    //FOO
        sop("procedureCall: " + ident.getText());
        #procedureCall = #([CALL, "CALL"], procedureCall);
      }
  ;

expression
  : multiplyingExpression ((PLUS^|MINUS^) multiplyingExpression)*
  ;

multiplyingExpression
  : signedExpression ((TIMES^|DIV^) signedExpression)*
  ;

signedExpression
  : (PLUS^|MINUS^)* primitiveElement
  ;

primitiveElement
  { String varx = ""; }
  : (IDENT LPAREN)=> procedureCall
  | varx=variableReference
  | constantValue
  | LPAREN! expression RPAREN!
  ;

variableReference returns [ String varx = ""]
  ///FOO { varx = ""; }
  : ident:IDENT
      { System.out.println("ident=  " + ident.getText());	//FOO
        varx = ident.getText();
      }
    ( LBRACKET! expression RBRACKET!
        { #variableReference = #([ARRAY, "ARRAY"], variableReference); }
    )?
  ;

constantValue
  : intlit:INTLIT
      { System.out.println("intlit= " + intlit.getText()); }	//FOO
  ;


//----------------------------------------------------------------------------
// The AssignLexer scanner
//----------------------------------------------------------------------------

class AssignLexer extends Lexer;

options {
    charVocabulary = '\0'..'\377';
    testLiterals = false;        // don't automatically test for literals
    k = 2;                       // two characters of lookahead
}


// Single-line comments
COMMENT
  : "//" (~('\n'|'\r'))*
    { $setType(Token.SKIP); }
  ;


// Literals
protected DIGIT : '0'..'9';

INTLIT : (DIGIT)+ ;


// Keywords are literals in the parser grammar

// Operators
SEMI     : ';' ;
COMMA    : ',' ;
EQUALS   : '=' ;
LBRACKET : '[' ;
RBRACKET : ']' ;
LPAREN   : '(' ;
RPAREN   : ')' ;
PLUS     : '+' ;
MINUS    : '-' ;
TIMES    : '*' ;
DIV      : '/' ;


// Whitespace -- ignored
WS
  : ( ' '
    | '\t'
    | '\f'

    // Handle newlines
    | ( "\r\n"  // Evil DOS
      | '\r'    // Macintosh
      | '\n'    // Unix (the right way)
      )
      { newline(); }
    )
    { $setType(Token.SKIP); }
  ;

// Multiple-line comments
ML_COMMENT
  : "/*"
    (    /* '\r' '\n' can be matched in one alternative or by matching
           '\r' in one iteration and '\n' in another.  I am trying to
            handle any flavor of newline that comes in, but the language
            that allows both "\r\n" and "\r" and "\n" to all be valid
            newline is ambiguous.  Consequently, the resulting grammar
            must be ambiguous.  I'm shutting this warning off.
        */
        options {
          generateAmbigWarnings = false;
        }
      :
        { LA(2)!='/' }? '*'
      |  '\r' '\n'       { newline(); }
      |  '\r'            { newline(); }
      |  '\n'            { newline(); }
      |  ~('*'|'\n'|'\r')
    )*
    "*/"
    { $setType(Token.SKIP); }
  ;


// An identifier.  Note that testLiterals is set to true.  This means
// that after we match the rule, we look in the literals table to see
// if it's a literal or really an identifer
IDENT
  options { testLiterals=true; }
  : ('a'..'z'|'A'..'Z') ('a'..'z'|'A'..'Z'|'0'..'9')*
  ;


//----------------------------------------------------------------------------
// The AssignTree tree-walker
// Perform elementary semantic analysis
//----------------------------------------------------------------------------

class AssignTreeWalker extends TreeParser;

options {
    importVocab = AssignParser;
}

/*
  These rule definitions all use the AST syntax:
    #(ROOT child.1 child.2 ... child.n);
*/

program
  { AST ast = null;}
  : #(PROGRAM (ast=expression[true])*)
  ;

expression [ boolean bTst ] returns [ AST r = _t ]
  { AST a, b = null; }

  : #(CALL id01:IDENT (a=expression[false])*
      { AssignParser.pSymbols.put(id01.getText(), 
                                  new Symbol(id01.getText(), Symbol.FUNCTION));
      }
    )

  | #(EQUALS a=expression[false] b=expression[true])

  | #(PLUS   a=expression[true] b=expression[true])

  | #(MINUS  a=expression[true] (b=expression[true])?)

  | #(TIMES  a=expression[true] b=expression[true])

  | #(DIV    a=expression[true] b=expression[true])

  | #(ARRAY  ida:IDENT a=expression[true])
      {
        if (bTst && AssignParser.pSymbols.check(ida.getText(), 
                    "ERROR: Use of an undefined array (" + ida.getText() + ")"))
            AssignParser.error();
        else
            AssignParser.pSymbols.put(ida.getText(), 
                                  new Symbol(ida.getText(), Symbol.ARRAYREF));
      }

  | id:IDENT
      { r = id;
        if (bTst && AssignParser.pSymbols.check(id.getText(),
                    "ERROR: Use of an identifier (" + id.getText() 
                                                      + ") that has no value"))
            AssignParser.error();
        else
            AssignParser.pSymbols.put(id.getText(), 
                                  new Symbol(id.getText(), Symbol.VARIABLE));
      }

  | num:INTLIT
      { r = num;
        AssignParser.pSymbols.put(num.getText(), 
                                  new Symbol(num.getText(), Symbol.INTEGER));
      }
  ;

// EOF
